import sys 
sys.path.append('.')
from cetl.utils.pipeline_v1 import DataPipeline


result=""
cfg = {"pipeline":[
                    {"type":"addRowNumber"},
                    {"type":"parallelTransformer", "isParallel":1, "transformers":[
                            {"type":"format2String", "description":"help following filterBy transformer"},
                            {"pipeline":[{"type":"filterBy"}]},
                            {"pipeline":[{"type":"paddingZero"}]},
                            {"type":"dropColumns"}
                        ]},
                    {"type":"passDataFrame"},
                    {"pipeline":[   {"type":"passDataFrame"},
                                    {"type":"toCSV", "out_dir":"./", "out_file":"out.csv"}]}
        ]}


pipe = DataPipeline(cfg)
print(pipe._node_pairs)
pipe = pipe.build_digraph()
# python3.6 cetl/tests/sample1.py
pipe.save_png("./sample1.1.png")
# print(pipe._dot)


from ..utils.builder import DataFrame, PANDAS_TRANSFORMERS
from ..utils.base import Base
from ..utils.transform_wrapper import transform_wrapper

@PANDAS_TRANSFORMERS.add()
class filterBy(Base):
    """
    {"type":"filterBy", "fieldname":"keep", "fieldvalue":"yes", "Op":"true"}
    """
    def __init__(self, fieldname=None, fieldvalue=None, Op="true", parallel_index=None):
        super().__init__()
        
        self.fieldname = fieldname
        self.fieldvalue = fieldvalue
        self.Op = Op
        self.parallel_index=parallel_index

    @transform_wrapper
    def transform(self, dataframe):
        #skip transformation if dataframe is an empty dataframe
        if dataframe.empty:
            return dataframe

        if isinstance(dataframe, list):
            assert isinstance(self.parallel_index, int)
            df=dataframe[self.parallel_index]
            # print(df)
        elif isinstance(dataframe, DataFrame):
            df = dataframe
        else:
            print("input should be pandas DataFrame or list of pandas DataFrames")

        # print(df.columns)
        # if "keep" in df.columns:
        #     print(self.Op)
        #     print(df[df[self.fieldname]==self.fieldvalue])
        if self.Op=="true":
            df = df[df[self.fieldname]==self.fieldvalue]
            # print(self.parallel_index)
        elif self.Op=="false":
            df = df[df[self.fieldname]!=self.fieldvalue]
            # print(self.parallel_index)
        
        return df

@PANDAS_TRANSFORMERS.add()
class filterValuesInList(Base):
    """
    {"type":"filterValuesInList", "fieldname":"keep", 
        "fieldvalues":["CANCELLED", "REPLACED"], "Op":"true"}
    """
    def __init__(self, fieldname=None, fieldvalues=None, Op=None):
        super().__init__()

        self.fieldname = fieldname
        self.fieldvalues = fieldvalues
        self.Op = Op

    @transform_wrapper
    def transform(self, dataframe):
        df = dataframe
        # print(df[self.fieldname])
        if self.Op == "true":
            c = df[self.fieldname].isin(self.fieldvalues)
            new_df = df[c]
            # print(new_df)
        else:
            c = df[self.fieldname].isin(self.fieldvalues)
            new_df = df[~(c)]
        return new_df




@PANDAS_TRANSFORMERS.add()
class filterByEvalStatement(Base):
    """
    {"type":"filterByEvalStatement", "eval_statement":""}
    """
    def __init__(self, 
                eval_statement=None):
        super().__init__()

        self.eval_statement = eval_statement

    @transform_wrapper
    def transform(self, dataframe):
        df = dataframe
        
        c = eval(self.eval_statement)
        df = df[c]


        return df



from datetime import timedelta
from airflow import DAG
from airflow.operators.python_operator import PythonOperator
from airflow operators.dummy_operator import DummyOperator
from pipeline_v2 import DataPipeline


cfg = {"pipeline":[

                    ],
        "dag_settings":{
            "dag_id":"kipling_cn",
            "schedule_interval":"@daily",
            "default_args":{"owner":"clement",
                            "retries":1,
                            "retry_delay":"1min",
                            "start_date":"2022-07-07"}
        }}

def build_dag(self, cfg, dag_settings):
    with DAG(
                dag_id = dag_settings["dag_id"],
                schedule_interval = dag_settings["schedule_interval"],
                default_args=dag_settings["default_args"],
                catchup=False) as dag:
        start = DummyOperator(task_id="start")

        end = DummyOperator(task_id="end")

        for tran_cfg  in cfg:
            






import numpy as np
from datetime import datetime
from ..utils.builder import DataFrame, PANDAS_TRANSFORMERS
from ..utils.base import Base
from ..utils.transform_wrapper import transform_wrapper
from ..utils.formating import addUniqueKeyList

@PANDAS_TRANSFORMERS.add()
class addNewColumn(Base):
    """This transformer is to add new column to the pandas dataframe

    Args:
        base_field (str):   Indicate the new column name to be added
        value (str): Indicate the value in the new column
    Returns:
        df (pandas.DataFrame): the data container type is dataframe

    Usage:
        {"type":"addNewColumn", "base_field":"Tier", "value":""}
    """
    def __init__(self, 
                base_field=None,
                value=None):
        super().__init__()
        self.base_field = base_field
        self.value = value

    @transform_wrapper
    def transform(self, dataframe: DataFrame):
        #skip transformation if dataframe is an empty dataframe
        if dataframe.empty:
            return dataframe

        df = dataframe
        df[self.base_field] = self.value
        
        return df

@PANDAS_TRANSFORMERS.add()
class addNewColumns(Base):
    """
    Usage:
        {"type":"addNewColumns", "base_fields":["Tier"], "values":[""]}
    """
    def __init__(self, 
                base_fields=None,
                values=None):
        super().__init__()
        self.base_fields = base_fields
        self.values = values

    @transform_wrapper
    def transform(self, dataframe: DataFrame) -> DataFrame:
        #skip transformation if dataframe is an empty dataframe
        if dataframe.empty:
            return dataframe

        df = dataframe

        for field, value in zip(self.base_fields, self.values):
                df[field] = value
        return df


@PANDAS_TRANSFORMERS.add()
class addNewColumnForIsWithinDayDiff(Base):
    """
    Usage:
        {"type":"addNewColumnForIsWithinDayDiff", "new_field":"", "date_field1":"", "date_field2":"", "day_limit":5}
    """
    def __init__(self, 
                new_field=None,
                date_field1=None,
                date_field2=None,
                day_limit=None):
        super().__init__()
        self.date_field1 = date_field1
        self.date_field2 = date_field2
        self.day_limit = day_limit
        self.new_field = new_field

        # if filter=="true":
        #     self.filter=True
        # else:
        #     self.filter=False

    @transform_wrapper
    def transform(self, dataframe: DataFrame) -> DataFrame:
        #skip transformation if dataframe is an empty dataframe
        if dataframe.empty:
            return dataframe

        df = dataframe

        df[self.new_field] = abs(df[self.date_field1]-df[self.date_field2])
        df[self.new_field] = df[self.new_field].dt.days<=self.day_limit

        # if self.filter:
        #     df = df[df[self.new_field]]

        return df



@PANDAS_TRANSFORMERS.add()
class addNewColumnForDayDiff(Base):
    """
    {"type":"addNewColumnForDayDiff", "new_field":"", "date_field1":"", "date_field2":""}
    """
    def __init__(self, 
                new_field=None,
                date_field1=None,
                date_field2=None):
        super().__init__()

        self.new_field = new_field
        self.date_field1 = date_field1
        self.date_field2 = date_field2

    @transform_wrapper
    def transform(self, dataframe: DataFrame) -> DataFrame:
        #skip transformation if dataframe is an empty dataframe
        if dataframe.empty:
            return dataframe

        df = dataframe

        df[self.new_field] = (df[self.date_field1]-df[self.date_field2]).dt.days


        return df


@PANDAS_TRANSFORMERS.add()
class addNewColumnWithNowDatetimeStr(Base):
    """
    {"type":"addNewColumnWithNowDatetimeStr", "new_field":"", "format":"'%Y-%m-%d %H:%M:%S'"}
    """
    def __init__(self, 
                new_field=None,
                format=None):
        super().__init__()

        self.new_field = new_field
        self.format = format

    @transform_wrapper
    def transform(self, dataframe: DataFrame) -> DataFrame:
        #skip transformation if dataframe is an empty dataframe
        if dataframe.empty:
            return dataframe

        df = dataframe

        # create now datetime string
        today = datetime.now()
        now_datetime_str = datetime.strftime(today, self.format)
        df[self.new_field] = now_datetime_str
        
        return df



@PANDAS_TRANSFORMERS.add()
class addNewColumnByPdSeries(Base):
    def __init__(self, 
                base_field=None,
                pd_series=None):

        super().__init__()
        self.base_field = base_field
        self.pd_series = pd_series

    @transform_wrapper
    def transform(self, dataframe: DataFrame) -> DataFrame:
        #skip transformation if dataframe is an empty dataframe
        if dataframe.empty:
            return dataframe

        df = dataframe
        df[self.base_field] = self.pd_series
        return df



@PANDAS_TRANSFORMERS.add()
class addNewColumnFromExisting(Base):
    """
    {"type":"addNewColumnFromExisting", "base_field":"", "new_field":""}
    """
    def __init__(self,
                 base_field=None,
                 new_field=None):

        super().__init__()
        self.base_field = base_field
        self.new_field = new_field
    

    @transform_wrapper
    def transform(self, dataframe: DataFrame) -> DataFrame:
        #skip transformation if dataframe is an empty dataframe
        if dataframe.empty:
            return dataframe

        df = dataframe
        df[self.new_field] = df[self.base_field]
        return df  



@PANDAS_TRANSFORMERS.add()
class addKeyColumnFromExisting(Base):
    def __init__(self,
                 subset=None,
                 new_field=None):
        super().__init__()

        self.subset = subset
        self.new_field = new_field

    @transform_wrapper
    def transform(self, dataframe: DataFrame) -> DataFrame:
        #skip transformation if dataframe is an empty dataframe
        if dataframe.empty:
            return dataframe

        df = dataframe
        headers = df.columns
        df[self.new_field]=""
        for field in self.subset:
            if field in headers:
                pd_series = df[field]
                df[self.new_field] = df[self.base_field] + "_" + pd_series.astype(str)
            else:
                print(f"no field named {field} in the dataframe")
        
        df[self.new_field] = df[self.new_field].apply(lambda x: x[1:])
        return df


@PANDAS_TRANSFORMERS.add()
class addNowColumn(Base):
    def __init__(self,
                 base_field=None,
                 now_format=None):
        super().__init__()

        self.base_field = base_field
        self.now_format = now_format


    @transform_wrapper
    def transform(self, dataframe: DataFrame) -> DataFrame:
        pass



@PANDAS_TRANSFORMERS.add()
class addNewFromConcatExistingColumns(Base):
    """"
    {"type":"addNewFromConcatExistingColumns", "subset":"", "new_fieldname":"", "delimiter":""}
    """
    def __init__(self,
                 subset=None,
                 new_fieldname=None,
                 delimiter=" "):
        super().__init__()

        self.subset = subset
        self.new_fieldname = new_fieldname
        self.delimiter=delimiter

    def concat_values(self, delimiter, values):
        if delimiter:
            new_value = delimiter.join(values)
        else:
            new_value = "".join(values)
        
        return new_value

    @transform_wrapper
    def transform(self, dataframe: DataFrame) -> DataFrame:
        df = dataframe

        headers = df.columns

        exist_headers = []
        df[self.new_fieldname]=""
        for field in self.subset:
            if field in headers:
                exist_headers.append(field)
            else:
                print(f"header {field} not exist in the dataframe")
                
        # concat method 2
        df[self.new_fieldname] = df[exist_headers[0]]+"_"+df[exist_headers[1]]
        # df[self.new_fieldname]=df.apply(lambda r: self.concat_values(self.delimiter, r[exist_headers]), axis=1)
        
        return df


@PANDAS_TRANSFORMERS.add()
class addUniqueKeyColumn(Base):
    """
    {"type":"addUniqueKeyColumn", "base_field":"", "value_prefix":"IMP1", "value_suffix":"", "value_start":1, "padding_digit":6}
    """
    def __init__(self,
                 base_field=None,
                 value_start=None,
                 value_prefix=None,
                 value_suffix=None,
                 padding_digit=None):
        super().__init__()

        self.base_field = base_field
        self.value_start = value_start
        self.value_prefix = value_prefix
        self.value_suffix = value_suffix
        self.padding_digit = padding_digit

    @transform_wrapper
    def transform(self, dataframe: DataFrame) -> DataFrame:
        df = dataframe
        

        value_end = df.shape[0]
        df[self.base_field] = addUniqueKeyList(value_prefix=self.value_prefix,
                                                value_suffix=self.value_suffix,
                                                value_start=self.value_start, 
                                                value_end = value_end, 
                                                padding_digit=self.padding_digit)

        return df



@PANDAS_TRANSFORMERS.add()
class addNewFromMultiplyCalculation(Base):
    """
    {"type":"addNewFromMultiplyCalculation", "new_field":"", 
            "multiply_fields":["Quantity", "Discount"]}
    """
    def __init__(self,
                 new_field=None,
                 multiply_fields=None):
        super().__init__()

        self.new_field = new_field
        self.multiply_fields = multiply_fields


    @transform_wrapper
    def transform(self, dataframe: DataFrame) -> DataFrame:
        df = dataframe

        headers = df.columns 

        # true fields checking
        true_fields = []
        for field in self.multiply_fields:
            if field in headers:
                true_fields.append(field)
            else:
                print(f"field {field} not exists")

        df[self.new_field] = 1
        
        for field in true_fields:
            df[self.new_field] = df[self.new_field] * df[field]

        return df


@PANDAS_TRANSFORMERS.add()
class addNewFromMinusCalculation(Base):
    """
    {"type":"addNewFromMinusCalculation", "new_field":"", 
            "minus_fields":["gap_to_next_grade", "Discount"]}
    """
    def __init__(self,
                 new_field=None,
                 minus_fields=None):
        super().__init__()

        self.new_field = new_field
        self.minus_fields = minus_fields

    @transform_wrapper
    def transform(self, dataframe: DataFrame) -> DataFrame:
        df = dataframe

        headers = df.columns 

        # true fields checking
        true_fields = []
        for field in self.minus_fields:
            if field in headers:
                true_fields.append(field)
            else:
                print(f"field {field} not exists")

        
        
        for i, field in enumerate(true_fields):
            if i ==0:
                df[self.new_field] = df[field]
            else:
                df[self.new_field] = df[self.new_field] - df[field]

        return df
        



@PANDAS_TRANSFORMERS.add()
class addMarkColumnFromWithinPeriod(Base):
    """
    {"type":"addMarkColumnFromWithinPeriod", "new_field":"isWithinGradePeriod", "period_start_field":"EffectiveDate", 
        "period_end_field":"ExpiryDate", "check_datetime_field":"MovementDate")
    """
    def __init__(self,
                 new_field=None,
                 period_start_field=None,
                 period_end_field=None,
                 check_datetime_field=None):
        super().__init__()

        self.new_field = new_field
        self.period_start_field = period_start_field
        self.period_end_field = period_end_field
        self.check_datetime_field = check_datetime_field

    @transform_wrapper
    def transform(self, dataframe) -> DataFrame:
        df = dataframe
        c1 = df[self.period_start_field]<=df[self.check_datetime_field]
        c2 = df[self.check_datetime_field]<=df[self.period_end_field]

        # df=None
        df[self.new_field] = "0"
        df.loc[(c1 & c2), self.new_field] = "1"
        return df



@PANDAS_TRANSFORMERS.add()
class addNewColumnWithReplaceDict(Base):
    """
    {"type":"addNewColumnWithReplaceDict", "new_field":"", "base_field":"", "values_dict":{}}
    """
    def __init__(self,
                new_field=None,
                base_field=None,
                values_dict=None):
        super().__init__()

        self.new_field = new_field
        self.base_field = base_field
        self.values_dict = values_dict

    @transform_wrapper
    def transform(self, dataframe) -> DataFrame:

        df = dataframe

        #create the new field
        df[self.new_field] = ""
        df[self.new_field] = df[self.base_field].replace(self.values_dict)

        return df

@PANDAS_TRANSFORMERS.add()
class addNewColumnWithDateAdjustment(Base):
    """
    {"type":"addNewColumnWithDateAdjustment", "new_field":"", "base_field":"", "year":-1, "month":0, "day":0}
    """
    def __init__(self,
                new_field=None,
                base_field=None,
                year=None,
                month=None,
                day=None):
        super().__init__()

        self.new_field = new_field
        self.base_field = base_field
        self.year = 0 if not year else year
        self.month = 0 if not month else month
        self.day = 0 if not day else day


    @transform_wrapper
    def transform(self, dataframe) -> DataFrame:

        df = dataframe

        #create the new field
        df[self.new_field] = df[self.base_field].apply(lambda x: x.replace(year=x.year+self.year, 
                                                    month=x.month+self.month,
                                                    day=x.day+self.day))
        return df

@PANDAS_TRANSFORMERS.add()
class addNewColumnWithDateTimeAdjustment(Base):
    """
    {"type":"addNewColumnWithDateAdjustment", "new_field":"", "base_field":"", "year":-1, "month":0, "day":0, "hour":23, "minute":59, "second":59}
    """
    def __init__(self,
                new_field=None,
                base_field=None,
                year=None,
                month=None,
                day=None,
                hour=None,
                minute=None,
                second=None):
        super().__init__()

        self.new_field = new_field
        self.base_field = base_field
        self.year = 0 if not year else year
        self.month = 0 if not month else month
        self.day = 0 if not day else day
        self.hour = 0 if not hour else hour
        self.minute = 0 if not minute else minute
        self.second = 0 if not second else second


    @transform_wrapper
    def transform(self, dataframe) -> DataFrame:

        df = dataframe


        def replace_datetime(x):

            # print(f"month:{x.month}, day:{x.day}")
            #avoid non leap year Feb 29
            cond1 = (x.year+self.year) % 4 >0
            cond2 = (x.month+self.month)==2
            cond3 = (x.day + self.day) == 29
            
            if cond1 and cond2 and cond3:
                return x.replace(year=x.year+self.year, 
                        month=x.month + self.month,
                        day=x.day + self.day - 1,
                        hour=x.hour + self.hour,
                        minute=x.minute + self.minute,
                        second=x.second + self.second)
            else:
                return x.replace(year=x.year+self.year, 
                        month=x.month + self.month,
                        day=x.day + self.day,
                        hour=x.hour + self.hour,
                        minute=x.minute + self.minute,
                        second=x.second + self.second)


        #create the new field
        df[self.new_field] = df[self.base_field].apply(lambda x: replace_datetime(x))
        return df



@PANDAS_TRANSFORMERS.add()
class addNewColumnFromGradeHistory(Base):
    """
    {"type":"addNewColumnFromGradeHistory", "sort_fields":["customer_id", "id"], "grouping_fields":["customer_id"], 
        "grade_history_id":"id", "new_field":"pre_grade_history_id", "is_ascending":"true", "na_position":"last"}

    Description: This method is going to help you find the pre-grade is of the same customer
    """
    def __init__(self,
                sort_fields=None,
                grouping_fields=None,
                grade_history_id=None,
                new_field=None,
                is_ascending=None,
                na_position="last"):
        super().__init__()

        self.sort_fields = sort_fields
        self.grouping_fields = grouping_fields
        self.grade_history_id = grade_history_id
        self.new_field = new_field
        if is_ascending=="true":
            self.is_ascending=True
        else:
            self.is_ascending=False

        self.na_position = na_position


    @transform_wrapper
    def transform(self, dataframe) -> DataFrame:

        df = dataframe

        # step 1: sort value by field of customer (i.e. customer_id)
        # if sort_fields is ["customer_id", "id"], then sorting is done from the right to the left
        # i.e. sort "id" first, then sort by "customer_id"
        df = df.sort_values(by=self.sort_fields, ascending=self.is_ascending, na_position=self.na_position)

        # step 2: get the list of grade history ids
        pre_grade_history_id_list = df[self.grade_history_id].to_list()
        pre_grade_history_id_list = [None] + pre_grade_history_id_list[:-1]

        # step 3: create the new field
        # print(pre_grade_history_id_list)
        # assert False
        df[self.new_field] = pre_grade_history_id_list

        # step 4: edit the first record of each group (customers)
        # print(f"df.groupby({self.grouping_fields})[{self.sort_fields}].head({1}).index")
        c = df.groupby(self.grouping_fields)[self.sort_fields].head(1).index
        df.loc[c, self.new_field] = None

        # format to int
        # df[self.new_field]=df[self.new_field].astype(int)

        return df



@PANDAS_TRANSFORMERS.add()
class addNewColumnFromEvalStatement(Base):
    """
    {"type":"addNewColumnFromEvalStatement", "new_field":"grade_value", "eval_value":"df[\"grade_id\"]-df[\"pre_grade_id\"]"}

    Description: This method is going to help you find the pre-grade is of the same customer
    """
    def __init__(self,
                new_field=None,
                eval_value=None):
        super().__init__()

        self.new_field = new_field
        self.eval_value = eval_value



    @transform_wrapper
    def transform(self, dataframe) -> DataFrame:

        df = dataframe

        df[self.new_field]=eval(self.eval_value)

        return df


@PANDAS_TRANSFORMERS.add()
class addGradeChangeType(Base):
    """
    {"type":"addGradeChangeType", "new_field":"temp_grade_change", "grade_diff":"grade_diff"}
    """
    def __init__(self, grade_diff=None, new_field=None):
        super().__init__()

        self.grade_diff=grade_diff
        self.new_field = new_field



    @transform_wrapper
    def transform(self, dataframe) -> DataFrame:
        df = dataframe

        # df[self.grade_diff].fillna(-800)

        c= (df[self.grade_diff]>0) & (df[self.grade_diff]<9)
        df.loc[c, self.new_field]="UPGRADE"

        c= (df[self.grade_diff]>=9)
        df.loc[c, self.new_field]="Set_Grade"
        
        c= df[self.grade_diff]<0
        df.loc[c, self.new_field]="DOWNGRADE"

        c= df[self.grade_diff]==0
        df.loc[c, self.new_field]="Renew"


        c = df[self.grade_diff].isnull()
        df.loc[c, self.new_field]="Opening Grade"

        
        return df

@PANDAS_TRANSFORMERS.add()
class addNewColumnForRowNumberToGroup(Base):
    """
    {"type":"addNewColumnForRowNumberToGroup", "group_fields":[], "new_field":""}
    """
    def __init__(self, 
                group_fields=None, 
                new_field=None):
        super().__init__()
        
        self.group_fields=group_fields
        self.new_field = new_field



    @transform_wrapper
    def transform(self, dataframe: DataFrame) -> DataFrame:
        #skip transformation if dataframe is an empty dataframe
        if dataframe.empty:
            return dataframe

        df = dataframe

        df[self.new_field] = df.groupby(self.group_fields).cumcount()+1
        
        return df


# from pyspark import SparkConf
from pyspark.sql import SparkSession
from pyspark.sql.functions import col, udf
from pyspark.sql.types import StringType
from cetl import pd
import pytz
import datetime

# SparkConf().set("spark.driver.memory", "50g")
# SparkConf().set('spark.executor.cores', '4')
# SparkConf().set('spark.cores.max', '4')

#maxToStringFields will affect df.show(10, truncate=False)
class SPARK_UTILS:
    def __init__(self, appName="test"):
        self.appName=appName

    def __enter__(self):
        self.spark = SparkSession \
                    .builder \
                    .appName(self.appName) \
                    .getOrCreate()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.spark.stop()

    def read_csv_with_pyspark(self, filepath, delimiter=","):
        sparkDF = self.spark.read.options(delimiter="|", header=True).csv(filepath)
        return sparkDF

    def convertPd2SparkDF(self, pd_df: pd.DataFrame):
        sparkDF = self.spark.createDataFrame(pd_df)
        return sparkDF

    def send2kafka(self, keyPairDF=None, connect_str=None, topic=None):
        """
        send2kafka(sparkDF, host="192.168.1.61:6667", topic="PYSPARK.campaign")
        """
        keyPairDF.selectExpr("CAST(key AS STRING)", "CAST(value AS STRING)") \
                .write \
                .format("kafka") \
                .option("kafka.bootstrap.servers", connect_str) \
                .option("topic", topic) \
                .save()


def format_scala_int(x):
    if not x:
        return 0

    if isinstance(x, int):
        return x
    elif isinstance(x, float):
        return int(x)
    elif isinstance(x, str):
        if x.isnumeric:
            return int(float(x))
        else:
            return x

def format_int(df, subset):
    headers = list(df.columns)
    for field in subset:
        if field in headers:
            #float(x) first due to int not able to deal with int('1.0')
            df[field]=df[field].apply(lambda x: format_scala_int(x))
        else:
            print(f'{field} column not exists')
    return df



def format_scala_float(x):
    if not x:
        return float(0)

    if isinstance(x, int):
        return float(x)
    elif isinstance(x, float):
        return x
    elif isinstance(x, str):
        if x.isnumeric:
            return float(x)
        else:
            return x

def format_float(df, subset):
    headers = list(df.columns)
    for field in subset:
        if field in headers:
            df[field] = df[field].apply(lambda x: format_scala_float(x))
        else:
            print(f"{field} column not exists")
    return df


def numeric_format(sparkDF, subset, data_type="int"):
    df = sparkDF

    headers = sparkDF.columns
    for field in subset:
        if field in headers:
            df=df.withColumn(field, col(field).cast(data_type))
            
    return df

def read_csv_from_sftp(sftp, filepath, delimiter=","):
    print(filepath)
    f = sftp.open(filepath)
    f.prefetch()
    df = pd.read_csv(f, delimiter=delimiter)
    return df


def convert2str(pd_df):
    df = pd_df
    headers = list(df.columns)
    for header in headers:
        df[header]=df[header].astype(str)
    return df

def add_index_column(pd_df, index_column_name="key"):
    df =pd_df
    df[index_column_name] = df.index + 1
    return df

def add_key_column(pd_df):
    df = pd_df
    # get datetime of today
    today = datetime.datetime.now()
    # conver the datetime to string
    now_datetime_str = datetime.datetime.strftime(today, '%Y-%m-%d %H:%M:%S')
    # concat the now_datetime_str and row index as unique
    # sometime the row index is not exists in the dataframe, so i regenereate
    indicesSeries = pd.Series([i+1 for i in range(df.shape[0])])
    df["key"] = now_datetime_str + "_" + indicesSeries

    return df

def convertPd2SparkDF(pd_df: pd.DataFrame):
    sparkDF = spark.createDataFrame(pd_df)
    return sparkDF


def send2kafka(keyPairDF=None, connect_str=None, topic=None):
    """
    send2kafka(sparkDF, host="192.168.1.61:6667", topic="PYSPARK.campaign")
    """
    keyPairDF.selectExpr("CAST(key AS STRING)", "CAST(value AS STRING)") \
            .write \
            .format("kafka") \
            .option("kafka.bootstrap.servers", connect_str) \
            .option("topic", topic) \
            .save()

########################################this method too slow
def pd_change_local_to_another_timezone(datetime_without_tz, origin_timezone, new_timezone):
    # print(datetime_without_tz)
    datetime_with_tz = origin_timezone.localize(datetime_without_tz, is_dst=None)
    return datetime_with_tz.astimezone(new_timezone)

def pd_change_timezone_to_millsecond(dataframe, subset, origin_timezone, new_timezone):

    df = dataframe

    origin_timezone = pytz.timezone(origin_timezone)
    new_timezone = pytz.timezone(new_timezone)

    for field in subset:
        df[field] = pd.to_datetime(df[field], format="%Y-%m-%d %H:%M:%S", errors="coerce")
        df[field]=df[field].apply(lambda x: pd_change_local_to_another_timezone(x,
                                                                            origin_timezone, 
                                                                            new_timezone))

        #change to millisecond
        df[field]=df[field].dt.strftime("%s")+"000"
    return df





def change_timezone_to_millsecond(dataframe, subset, origin_timezone, new_timezone):
    # origin_timezone = "Asia/Hong_Kong"
    # new_timezone = "UTC"
    origin_timezone = pytz.timezone(origin_timezone)
    new_timezone = pytz.timezone(new_timezone)

    df = dataframe

    #customer function
    def change_local_to_another_timezone(datetime_without_tz):
        # print(datetime_without_tz)
        if datetime_without_tz:
            datetime_without_tz = pd.to_datetime(datetime_without_tz, format="%Y-%m-%d %H:%M:%S", errors="coerce")
            datetime_with_tz = origin_timezone.localize(datetime_without_tz, is_dst=None)
            #change the timezone
            t = datetime_with_tz.astimezone(new_timezone)
            #time to str
            t_str = datetime.datetime.strftime(t, '%s') + "000"
            #return cast to int
            return int(t_str)
        else:
            return datetime_without_tz

    udf_change_local_timezone = udf(change_local_to_another_timezone)

    for field in subset:
        df = df.withColumn(field, udf_change_local_timezone(col(field)))

    return df


def replace_na_value_dict(target_dict):
    target_dict = {k:"" if not v else v for k, v in target_dict.items()}
    return target_dict
def replace_str_dict_na_value(str_dict):
    print(type(str_dict))
    str_dict = str_dict.replace("null", "\"\"")
    return str_dict




def udf_replace_keyvaluepair_value_col_null(keypair_DF):
    udf_replace_str_dict_na_value = udf(replace_str_dict_na_value)
    keypair_DF = keypair_DF.withColumn("value", udf_replace_str_dict_na_value(col("value")))
    return keypair_DF


def udf_string_null_format(sparkDF, subset=None):
    def replace_value(str_value):
        if not str_value:
            return ""

        return str_value
    udf_replace_value = udf(replace_value)
    headers = sparkDF.columns
    for field in subset:
        if field in headers:
            sparkDF = sparkDF.withColumn(field, udf_replace_value(col(field)))
            if sparkDF.schema[field].dataType==StringType():
                sparkDF = sparkDF.withColumn(field, udf_replace_value(col(field)))
            else:
                print(f'   column {field} is not string type')
        else:
            print(f'   column {field} is not valid')
    return sparkDF


    new_sparkDF.dtypes


import numpy as np

def generate_batches(samples_in_batch=100, items_list=None):
    population=len(items_list)
    batches =[]
    batch=[]
    for i in np.arange(population):
        if len(batch) < samples_in_batch-1:
            batch.append(items_list[i])
        else:
            batch.append(items_list[i])
            batches.append(batch)
            batch=[]

        if i == population-1 and len(batch)>0:
            batches.append(batch)
        
    return batches


from ..utils.builder import DataFrame, PANDAS_TRANSFORMERS
from ..utils.base import Base
from ..utils.transform_wrapper import transform_wrapper

@PANDAS_TRANSFORMERS.add()
class addRowNumber(Base):
    """
    {"type":"addRowNumber", "mark_field":"rownumber"}
    """
    def __init__(self, mark_field="Number"):
        super().__init__()
        
        self.mark_field=mark_field

    @transform_wrapper
    def transform(self, dataframe: DataFrame) -> DataFrame:
        #skip transformation if dataframe is an empty dataframe
        if dataframe.empty:
            return dataframe

        df = dataframe
        df['A'] = "A"
        df[self.mark_field] = df.groupby(['A']).cumcount()+1
        df=df.drop(columns=['A'])
        return df





from ..utils.builder import DataFrame, PANDAS_TRANSFORMERS
from ..utils.base import Base
from ..utils.transform_wrapper import transform_wrapper

@PANDAS_TRANSFORMERS.add()
class addHeaders2Dataframe(Base):
    def __init__(self, headers=None):
        self.headers=headers



    @transform_wrapper
    def transform(self, dataframe: DataFrame):
        #skip transformation if dataframe is an empty dataframe
        if dataframe.empty:
            return dataframe
            
        df = dataframe
        # df = df.rename(columns={'oldName1': 'newName1', 'oldName2': 'newName2'})
        df.columns = self.headers
        # print(df)

        return df


@PANDAS_TRANSFORMERS.add()
class renameDataframeHeaders(Base):
    """
    {"type":"renameDataframeHeaders", "headers_dict":{}}
    """
    def __init__(self, headers_dict=None):
        self.headers_dict = headers_dict



    @transform_wrapper
    def transform(self, dataframe:DataFrame):
        if dataframe.empty:
            return dataframe

        df = dataframe
        df = df.rename(self.headers_dict, axis=1)

        return df



@PANDAS_TRANSFORMERS.add()
class selectOneDataFrameFromMultiple(Base):
    """
    {"type":"selectOneDataFrameFromMultiple", "position":1}
    """
    def __init__(self, position=None):
        self.position = position



    @transform_wrapper
    def transform(self, dataframes:list):

        df = dataframes[self.position]

        return df


@PANDAS_TRANSFORMERS.add()
class addOneMoreDataFrameFromExistingDataframe(Base):
    """
    {"type":"addOneMoreDataFrameFromExistingDataframe", "subset":[]}
    """
    def __init__(self, subset=None):
        self.subset = subset



    @transform_wrapper
    def transform(self, dataframe):

        df = dataframe[self.subset]

        return [dataframe, df]


from ..utils.builder import DataFrame, FUNCTIONAL_TRANSFORMERS
from ..utils.base import Base
from ..utils.transform_wrapper import transform_wrapper

@FUNCTIONAL_TRANSFORMERS.add()
class dummyTransformer(Base):
    """
    Usage:
        if there is not found transformer in the cfg, the DataPipeline will use this dummy transformer to help
        {"type":"dummyTransformer", "data_container_type":"functional"}
    """
    def __init__(self):
        super().__init__()
        pass

    @transform_wrapper
    def transform(self, input_data):
        return input_data


import sys 
sys.path.append('.')
from cetl import DataPipeline

cfg = {"pipeline":[ {"type":"dummyStart", "module_type":"functional"},
                    {"type":"parallelTransformer", "transformers":[
                        {"breakpoint":1, "type":"generateDataFrame"},
                        {"type":"generateDataFrame"},
                        {"type":"generateDataFrame"}
                    ]},
                    {"type":"unionAll"}],
                    
        "pipeline_settings":{"print_cfg":1, "print_task_result":1}}

pipe = DataPipeline(cfg)
df = pipe.transform("")
# print(df)


pipe = pipe.build_digraph()
pipe.save_png("./sample7.png")


# python cetl/dev_tests/sample7_build_pipeline_from_cfg.py


from ..utils.base import Base
from ..utils.builder import JSON_TRANSFORMERS
from ..utils.transform_wrapper import transform_wrapper
import csv
import json

@JSON_TRANSFORMERS.add()
class readCSV(Base):
    def __init__(self,
                filepath=None , 
                delimiter=","):
        """
        the value of "keep" can be "last", False, "first"
        {"type":"readCSVAsJson", "filepath":"/home/clement/data/data_warehouse/Orders.csv", 
            "delimiter":","}
        """
        
        self.filepath = filepath
        self.delimiter=delimiter

    @transform_wrapper
    def transform(self, input_data):

        data = []
        with open(self.filepath, encoding='utf-8') as f:
            csvReader = csv.DictReader(f, delimiter=self.delimiter)
            # csvReader = csv.reader(f, delimiter=self.delimiter)
        
            # data = list(csvReader)
            for row in csvReader:
                data.append(row)

        json_str = json.dumps({"json_data":data}, indent=4)

        json_obj = json.loads(json_str)

        return json_obj


from email import header
import numpy as np
from ..utils.builder import pd, DataFrame, PANDAS_TRANSFORMERS
from ..utils.base import Base
from ..utils.transform_wrapper import transform_wrapper

@PANDAS_TRANSFORMERS.add()
class unionAll(Base):
    """
    {"type":"unionAll"}
    """
    def __init__(self):
        super().__init__()
        pass
        

    @transform_wrapper
    def transform(self, dataframes: list):
        #skip transformation if dataframe is an empty dataframe
        for dataframe in dataframes:
            if dataframe is None:
                return dataframes
            if dataframe.empty:
                return dataframes
        
        df = pd.concat(dataframes)
        return df


@PANDAS_TRANSFORMERS.add()
class sqlJoin(Base):
    """
    {"type":"sqlJoin", "how":"left", "left_column":"MemberID", "right_column":"MemberID"}
    """
    def __init__(self,
                how=None,
                left_column=None,
                right_column=None):
        super().__init__()

        self.how = how
        self.left_column = left_column
        self.right_column = right_column
        


    @transform_wrapper
    def transform(self, dataframes):
        #https://www.statology.org/pandas-merge-on-different-column-names/
        dataframe1 = dataframes[0]
        dataframe2 = dataframes[1]

        if dataframe1.empty or dataframe2.empty:
            return pd.DataFrame()

        df = pd.merge(dataframe1, 
                      dataframe2, 
                      how=self.how, 
                      left_on=self.left_column,
                      right_on = self.right_column)
        return df
    

class multiJoin(Base):
    """
    {"type":"multiJoin", rules=[{"how":"left", "left_column":"grade_id", "right_column":""},
                                {"how":"left", "left_column":"grade_id", "right_column":""},
                                {"how":"left", "left_column":"grade_id", "right_column":""}
                                ]}
    usage of reduce lambda
    -----------------------------
    from functools import reduce
    numbers = [0, 1, 2, 3, 4]

    def my_add(a, b):
        result = a + b
        print(f"{a} + {b} = {result}")
        return result

    reduce(my_add, numbers)
    0 + 1 = 1
    1 + 2 = 3
    3 + 3 = 6
    6 + 4 = 10
    """


    def __init__(self, rules=None):
        super().__init__()

        self.rules = rules

    @transform_wrapper
    def transform(self, input_list):
        from functools import reduce
        def newSqlJoin(first_cfg=None, second_cfg=None):
            left_df = first_cfg["df"]
            how=first_cfg["how"]
            left_column=first_cfg["left_column"]
            right_column=first_cfg["right_column"]
            right_df = second_cfg["df"]

            transformer = sqlJoin(how=how, left_column=left_column, right_column=right_column)
            df = transformer.transform([left_df, right_df])
            second_cfg["df"]=df
            return second_cfg
        
        # prepare iterrable list
        self.rules.append({})
        # every rule is a dict
        for rule, df in zip(self.rules, input_list):
            rule["df"]=df
        
        result = reduce(lambda cfg1, cfg2:newSqlJoin(cfg1, cfg2), self.rules)
        return self.output_df




import numpy as np
na_values = ["", 
            "#N/A", 
            "#N/A N/A", 
            "#NA", 
            "-1.#IND", 
            "-1.#QNAN", 
            "-NaN", 
            "-nan", 
            "1.#IND", 
            "1.#QNAN", 
            "<NA>", 
            "N/A", 
    #              "NA", 
            "NULL", 
            "NaN", 
            "n/a", 
            "nan", 
            "null",
            np.nan,
            float('nan'),
            np.float("nan"),
            np.float64("nan")]

def paddingzero(x, num_digit):
    if type(x)==str:
        return x.zfill(num_digit)
    else:
        type(x)==int
        # print("i am int")
        new_x = str(x).zfill(num_digit)
        # print(new_x)
        return new_x


def addUniqueKeyList(value_prefix = None,
                    value_suffix = None,
                    value_start = None,
                    value_end = None,
                    padding_digit=None):

    unique_list = [value_prefix + paddingzero(i, padding_digit) + value_suffix for i in range(value_start,value_end+1)]

    return unique_list


def format_int(df, subset):
    headers = list(df.columns)
    for field in subset:
        if field in headers:
            df[field] = df[field].apply(lambda x: int(x))
        else:
            print(f"{field} column not exists")



from ..utils.builder import DataFrame, PANDAS_TRANSFORMERS, pd
from ..utils.base import Base
from ..utils.transform_wrapper import transform_wrapper

import csv
import json

@PANDAS_TRANSFORMERS.add()
class generateDataFrame(Base):
    """
    {"type":"generateDataFrame"}
    """
    def __init__(self, data=None):
        super().__init__()
        self.data = data
    
    @transform_wrapper
    def transform(self, input) -> DataFrame:
        
        df = None
        if not self.data:
            self.data = [  {"customer_id":111, "first_name":"peter", "last_name":"Hong", "title":"Mr."},
                                {"customer_id":222, "first_name":"YuCheung", "last_name":"Wong", "title":"Mr."},
                                {"customer_id":333, "first_name":"Cindy", "last_name":"Wong", "title":"Mrs."},
                            ]
            df = pd.DataFrame(self.data)

        return df




from ..utils.builder import DataFrame, FUNCTIONAL_TRANSFORMERS
from ..utils.base import Base
from ..utils.transform_wrapper import transform_wrapper


@FUNCTIONAL_TRANSFORMERS.add()
class passDataFrame(Base):
    """do nothing
    {"type":"passDataFrame", "data_container_type":"functional"}
    """
    def __init__(self):
        super().__init__()
        pass

    @transform_wrapper
    def transform(self, input_data):
        return input_data


@FUNCTIONAL_TRANSFORMERS.add()
class dummyStart(Base):
    """do nothing
    {"type":"dummyStart", "data_container_type":"functional"}
    """
    def __init__(self):
        super().__init__()
        pass

    @transform_wrapper
    def transform(self, input):
        return ""



from sklearn.base import BaseEstimator, TransformerMixin

class Base(BaseEstimator, TransformerMixin):
    def __init__(self):
        self.node_id = ""
        self.node_name = ""
        self.description = ""
        self.isParallel = "false"
        self.total_time = 0
        self.total_input = 0
        self.total_output = 0
        self.data_container_type="pandas"
        self.start_datetime = ""
        self.end_datetime = ""
        self.breakpoint = 0
        self.print_task_result=0


    def fit(self):
        pass
    


from pyspark.sql import DataFrame, functions as F
import functools
from joblib import Parallel, delayed
from ..utils.builder import FUNCTIONAL_TRANSFORMERS
from ..utils.base import Base


@FUNCTIONAL_TRANSFORMERS.add()
class multiThreadTransformers(Base):
    
    def __init__(self, 
                 steps: list,
                 n_jobs=-1,
                 multithread=False):
        
        self.steps = steps
        self.n_jobs = n_jobs
        self.multithread = multithread
        
    def fit(self):
        pass
    

    def transform(self, X):

        assert len(self.steps)>0
        
        dfs=None
        if isinstance(X, list):
            X=X
        else:
            # prepare multiple input for parallel transform
            X = [X.copy(deep=True) if isinstance(X, DataFrame) else X for result in range(len(self.steps))]

            
        assert len(X) == len(self.steps)
        parallel_result = None

        if self.multithread:

            with Parallel(n_jobs=self.n_jobs, verbose=5) as parallel:
                funcs = [delayed(lambda t:t.transform(x))(transformer) for x, transformer in zip(X, self.steps)]
                parallel_result = parallel(funcs)
        else:
            parallel_result = [transformer.transform(x) for x, transformer in zip(X, self.steps)]

        return parallel_result


import sys 
sys.path.append('.')
from cetl.utils.pipeline_v1 import DataPipeline


result=""
cfg = {"pipeline":[
                    {"type":"addRowNumber"},
                    {"type":"parallelTransformer", "isParallel":1, "transformers":[
                            {"type":"format2String", "description":"help following filterBy transformer"},
                            {"pipeline":[   {"type":"filterBy"},
                                            {"type":"parallelTransformer", "isParallel":1, "transformers":[
                                                {"type":"addRowNumber"},
                                                {"type":"passDataFrame"}]},
                                            {"type":"toCSV", "out_dir":"./", "out_file":"out.csv"}]},
                            {"pipeline":[{"type":"paddingZero"}]},
                            {"type":"dropColumns", "description":"remove temp columns"}
                        ]},
                    {"type":"passDataFrame"},
                    {"pipeline":[   {"type":"passDataFrame"},
                                    {"type":"toCSV", "out_dir":"./", "out_file":"out.csv"}]}
        ]}


pipe = DataPipeline(cfg)
# print(pipe._node_pairs)
# print(pipe._nodename2transformer)
pipe = pipe.build_digraph()
pipe.save_png("./sample1.2.png")
# print(pipe._dot)


# python3.6 cetl/tests/sample2.py


from ..utils.builder import FUNCTIONAL_TRANSFORMERS
from ..utils.base import Base
from ..utils.transform_wrapper import transform_wrapper
import copy
from ..utils.file_mgt import get_datetime_str

@FUNCTIONAL_TRANSFORMERS.add()
class parallelTransformer(Base):
    def __init__(self, transformers: list):
        super().__init__()
        self.transformers = transformers


    def transform(self, X)-> list:
        if isinstance(X, list):
            X=X
        elif X is None:
            X = ""
            X = [X for i in range(len(self.transformers))]
        else:
            X = [copy.deepcopy(X) for i in range(len(self.transformers))]

        # print(X)
        # assert len(X)>0
        outputs = []
        for i, transformer in enumerate(self.transformers):
            if hasattr(transformer, "breakpoint"):
                if transformer.breakpoint:
                    # print(transform.node_name, "has breakpoint", transform.breakpoint)
                    outputs.append(transformer.transform(X[i]))
                    datetime_str = get_datetime_str(format="%Y-%m-%d %H:%M:%S")
                    print(f"{datetime_str} stop with breakpoint at {transformer.node_name}") if transformer.print_task_result else ""
                else:
                    outputs.append(transformer.transform(X[i]))    
            else:
                outputs.append(transformer.transform(X[i]))
            
        return outputs




class Registry:
    def __init__(self,  name=None):
        self._name = name
        self._module_dict= dict()

    @property
    def name(self):
        return self._name

    @property
    def module_dict(self):
        return self._module_dict

    def add(self, module_name=None, forece=False, module=None):
        """
        if there is input module, then will return module
        for example: add('Conv1d', module=nn.Conv1d)
        if there is no module, then will return _add
        """

        if module is not None:
            if module_name is None:
                module_name = module.__name__
            self._module_dict[module_name]= module
            return module

        def _add(inputClass):
            module_name = inputClass.__name__
            self._module_dict[module_name] = inputClass
            # print(module_name)
            # print("{} added".format(name))
            return inputClass
        return _add

    def get(self, key):
        return self._module_dict.get(key)

    def __len__(self):
        return len(self._module_dict)

    def __contains__(self, key):
        """
        help to make the class, "Registry" iterable
        """
        return self.get(key) is not None


    def __repr__(self):
        format_str = self.__class__.__name__ + \
                     f'(name={self._name}, ' \
                     f'items={self._module_dict})'
        return format_str


from .utils import *
from .utils.builder import (FUNCTIONAL_TRANSFORMERS, JSON_TRANSFORMERS, TEST_TRANSFORMERS, 
                            PANDAS_TRANSFORMERS, SPARK_TRANSFORMERS, DB_MAPPERS, DB_MODELS,
                            pd, DataFrame, context_name)
from .utils.pipeline import Pipeline, DataPipeline, make_pipeline, build_pipeline
from .utils.transform_wrapper import transform_wrapper
from .utils.base import Base

from .functional_modules import *
from .pandas_modules import *
# from .spark_modules import *
from .json_modules import *





from datetime import datetime
from distutils.log import error
import fnmatch
import os
import paramiko
from cetl.utils.builder import pd
import json
from cetl.utils.file_mgt import get_temp_name, get_file_ext

class SFTP:
    def __init__(self, 
                host=None, 
                port=22, 
                username=None, 
                password=None):

        self.host=host
        self.port=port
        self.username=username
        self.password = password

    def __enter__(self, ):
        tran = paramiko.Transport((self.host, int(self.port)))
        tran.connect(username=self.username, password=self.password)
        self.transport = tran

        try:        
            sftp = paramiko.SFTPClient.from_transport(self.transport)
            self.sftp = sftp
            return self
        except:
            self.transport.close()
            raise

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.sftp.close()
        self.transport.close()

    def search_file_in_sftp(self, target_dir, template_str):
        temp=[]
        for filename in self.sftp.listdir(target_dir):
            # print("#############", template_str)
            # print("              ", filename)
            if fnmatch.fnmatch(filename, template_str):
                temp.append(os.path.join(target_dir, filename))
        return temp

    def search_source_files(self, target_dir=None, template_strs=None):
        filepaths = []
        for template_str in template_strs:
            if not template_str:
                filepaths.append("")
            else:
                paths = self.search_file_in_sftp(target_dir, template_str)
                paths.sort()
                if len(paths)>0:
                    filepaths.append(paths[-1])#we want to get the latest one
                else:
                    filepaths.append("")

        # print(f'find paths: ', filepaths)
        return filepaths


    def search_range_of_files(self, target_dir=None, start=None, end=None, file_filter_regex=None):
        
        filenames = self.sftp.listdir(target_dir)

        # filter the filenames with target file_ext
        filter_filenames = []
        for filename in filenames:
            # file_filter_regex is like "my*.csv"
            if fnmatch.fnmatch(filename, file_filter_regex):
                filter_filenames.append(filename)

        # sort the filenames
        # https://www.geeksforgeeks.org/python-how-to-sort-a-list-of-strings/
        filter_filenames.sort()

        # get the position range
        if start=="":
            start_pos = 0
        else:
            start_pos = filter_filenames.index(start)

        if end=="":
            end_pos = -1
        else:
            end_pos = filter_filenames.index(end)+1
        filter_filenames = filter_filenames[start_pos:end_pos]
        filter_filepaths = [os.path.join(target_dir, filename) for filename in filter_filenames]
        all_filepaths = [os.path.join(target_dir, filename) for filename in filter_filenames]


        return all_filepaths, filter_filepaths


    def is_file_exists(self, filepath):
        target_dir = os.path.dirname(filepath)
        filename = os.path.basename(filepath)
        filepaths = self.search_file_in_sftp(target_dir, filename)
        if len(filepaths)>0:
            return True
        else:
            return False

    def remove_file(self, filepath):
        self.sftp.remove(filepath)

    def move_file(self, filename, from_dir, to_dir, overwrite=True):
        from_path = os.path.join(from_dir, filename)
        to_path = os.path.join(to_dir, filename)
        if self.is_file_exists(to_path) and overwrite:
            self.remove_file(to_path)
        self.sftp.rename(from_path, to_path)

    def read_csv_from_sftp(self, filepath, delimiter=",", header=0):
        # open the csv file
        with self.sftp.open(filepath) as f:
            f.prefetch()
            # print(header, "##########################################")
            df = pd.read_csv(f, delimiter=delimiter, header=header)
            return df

    def read_excel_from_sftp(self, 
                            filepath, 
                            delimiter=",", 
                            header=0,
                            skiprow_start=1,
                            skiprow_end=-1,
                            sheet_name=None, 
                            column_dtype=None):
                            

        with self.sftp.open(filepath) as f:
            f.prefetch()
            #dtype=column_dtype,
            df = pd.read_excel(f, sheet_name=sheet_name, index_col=None, header=header, skiprows = range(skiprow_start-1, skiprow_end))
            return df

    
    def read_json_as_dict(self, filepath):
        # open the jso nfile
        with self.sftp.open(filepath) as f:
            f.prefetch()
            json_data = json.load(f)
            return json_data


    def write_csv_to_sftp(self, filepaths, dfs, sep=",", header=None):
        for filepath, df in zip(filepaths, dfs):
            w = self.sftp.open(filepath, "w")
            w.write(df.to_csv(index=False, sep=sep, header=header))
    
    def write_temp_csv_to_sftp(self,
                                out_dir=None, 
                                file_prefix_list=None,
                                file_suffix_list=None,
                                dfs=None,
                                sep=None, 
                                header=True):

        filepaths= []
        for file_prefix, file_suffix in zip(file_prefix_list, file_suffix_list):
            out_file = os.path.join(out_dir, get_temp_name(file_prefix, file_suffix))
            filepaths.append(out_file)
        
        self.write_csv_to_sftp(filepaths, dfs, sep, header)

        return filepaths

    def mkdir(self, remote_directory):
        dir_path = str()
        for dir_folder in remote_directory.split("/"):
            if dir_folder == "":
                continue
            dir_path += r"/{0}".format(dir_folder)
            try:
                self.sftp.listdir(dir_path)
            except IOError:
                self.sftp.mkdir(dir_path)


def search_file_in_sftp(sftp, target_dir, template_str):
        temp=[]
        for filename in sftp.listdir(target_dir):
                if fnmatch.fnmatch(filename, template_str):
                       temp.append(os.path.join(target_dir, filename))
        return temp

def is_file_exists(sftp, filepath):
    target_dir = os.path.dirname(filepath)
    filename = os.path.basename(filepath)
    filepaths = search_file_in_sftp(sftp, target_dir, filename)
    if len(filepaths)>0:
        if filepaths[0]==filepath:
            return True
        else:
            return False
    else:
        return False



def remove_file(sftp, filepath):
    sftp.remove(filepath)

def move_file(sftp, filename, from_dir, to_dir, overwrite=True):
    from_path = os.path.join(from_dir, filename)
    to_path = os.path.join(to_dir, filename)
    if is_file_exists(sftp, to_path):
        remove_file(sftp, to_path)
    sftp.rename(from_path, to_path)


from .input import generateDataFrame
from .add_row_number import addRowNumber
from .add_new_column import (addNewColumn, addNewColumns, addNewColumnByPdSeries, addNewColumnFromExisting, 
                            addKeyColumnFromExisting, addUniqueKeyColumn, addNewFromConcatExistingColumns,
                            addNewFromMultiplyCalculation, addNewColumnWithNowDatetimeStr, 
                            addNewColumnWithDateAdjustment, addNewColumnWithDateTimeAdjustment, addNewColumnFromGradeHistory,
                            addNewColumnForIsWithinDayDiff, addNewColumnForDayDiff, addNewColumnForRowNumberToGroup,
                            addNewColumnFromEvalStatement, addGradeChangeType, addNewFromMinusCalculation,
                            addNewColumnWithReplaceDict)
from .filter import filterBy, filterValuesInList, filterByEvalStatement
from .two2one import unionAll, sqlJoin


from cetl.pandas_modules.input import generateDataFrame
from cetl import DataPipeline, pd


def test_pass_dataframe():

    answer = generateDataFrame().transform("")
    # print(answer)

    cfg = {"pipeline":[ {"type":"generateDataFrame"},
                        {"type":"passDataFrame", "data_container_type":"functional"}
                        ]}

    pipe = DataPipeline(cfg)
    df = pipe.transform("")

    assert df.equals(answer)



# def test_add_numbers():
#     assert add_numbers(2, 3) == 5
#     assert add_numbers(-1, 1) == 0
#     assert add_numbers(0, 0) == 0


from cetl import DataPipeline, pd
from cetl.pandas_modules.input import generateDataFrame

def test_parallel_transformer():

    answer = pd.concat([generateDataFrame().transform(""), 
                        generateDataFrame().transform("")])
    # print(answer)

    cfg = {"pipeline":[
                        {"type":"parallelTransformer", "transformers":[
                            {"type":"generateDataFrame"},
                            {"type":"generateDataFrame"}
                        ]},
                        {"type":"unionAll"}
    ]}

    pipe = DataPipeline(cfg)
    df = pipe.transform("")
    # print(df)

    assert df.equals(answer)


def test_parallel_transformer2():

    answer = pd.concat([generateDataFrame().transform(""), 
                        generateDataFrame().transform("")])
    # print(answer)

    cfg = {"pipeline":[ {"type":"dummyStart", "data_container_type":"functional"},
                        {"type":"parallelTransformer", "transformers":[
                            {"type":"generateDataFrame"},
                            {"type":"generateDataFrame"}
                        ]},
                        {"type":"unionAll"}
    ]}

    pipe = DataPipeline(cfg)
    df = pipe.transform("")
    # print(df)

    assert df.equals(answer)


import sys 
sys.path.append('.')
from cetl import DataPipeline



result=""
cfg = {"pipeline":[{"type":"readCSV", "data_container_type":"json",
                        "filepath":"/home/clement/data/for_json_pipeline/Orders.csv", 
                        "delimiter":"\t"},
                    {"type":"dropColumns", "data_container_type":"json", "subset":["EmployeeID", "ShipperID"]}
        ]}


pipe = DataPipeline(cfg)
for step in pipe.steps:
    print(type(step[1]))

# pipe = pipe.build_digraph()
# pipe.save_png("./sample5.png")


# python3.6 cetl/tests/sample5_data_container_type.py


from sklearn.pipeline import Pipeline as sklearn_pipeline
from collections import defaultdict
import copy
from flask import send_file
import tempfile
import os
import re
from cetl.functional_modules.parallel_transformer import parallelTransformer
from cetl.utils.builder import (build_transformer_from_cfg, get_register, FUNCTIONAL_TRANSFORMERS)
from .dot_node import update_node, recursive_build_digraph, add_dot_node, init_digraph, append_node_text
from .file_mgt import get_datetime_str


def _name_estimators(estimators):
    """Generate names for estimators."""

    names = [
        estimator if isinstance(estimator, str) else type(estimator).__name__.lower()
        for estimator in estimators
    ]
    namecount = defaultdict(int)
    for est, name in zip(estimators, names):
        namecount[name] += 1

    for k, v in list(namecount.items()):
        if v == 1:
            del namecount[k]

    for i in reversed(range(len(estimators))):
        name = names[i]
        if name in namecount:
            names[i] += "-%d" % namecount[name]
            namecount[name] -= 1

    return list(zip(names, estimators))
        
def make_pipeline(*steps, memory=None, verbose=False):
    """
    Examples
    -----------
    >>> from core.utils.pipeline import build_pipeline
    >>> build_pipeline(addNewColumn(), drop_columns())
    Pipeline(transformers=[ 'addNewColumn', addNewColumn(),
                            'drop_columns', drop_columns()])
    """
    # print(steps)
    return Pipeline(_name_estimators(steps), memory=memory, verbose=verbose)



def recursive_build_pipeline_from_cfg(  cfg, 
                                        node_name_set, 
                                        node_pairs, 
                                        latest_node, 
                                        index, 
                                        parallel_index, 
                                        nodename2transformer, 
                                        print_task_result):
    """
    Examples
    -----------
    global_index = 0
    latest_node = ""
    node_pairs = []

    pipeline, node_pairs, latest_node, index = recursive_build_pipeline_from_cfg(cfg, node_pairs, latest_node, global_index)
    """

    assert "pipeline" in cfg
    
    steps = []

    for trans_cfg in cfg["pipeline"]:
        trans_name = trans_cfg.get("type")  if "type" in trans_cfg else ""
        trans_name = "pipeline" if "pipeline" in trans_cfg else trans_name
        breakpoint = trans_cfg.pop("breakpoint") if "breakpoint" in trans_cfg else 0

        
        if trans_name == "parallelTransformer":         
            
            ########################## adding intermediate node for parallelTransformer
            parallel_name = "parallel"
            parallel_index +=1
            default_parallel_label = "parallelTransformer"
            parallel_node_name = f"{parallel_name} #{parallel_index}"
            

            transformers_list = []
            parallel_latest_node = []
            parallel_node_pairs = {}
            latest_nodes = None
            if isinstance(latest_node,str):
                #convert the _previous_node to a list
                latest_nodes= [latest_node for i in range(len(trans_cfg["transformers"]))]
            elif isinstance(latest_node, list):
                latest_nodes = latest_node
            else:
                assert False, "latest_node is not str or list"

            parallel_dict_index = [i for i in range(len(latest_nodes))]
            # print(parallel_dict_index, latest_nodes, trans_cfg)
            for i, latest_node, sub_trans_cfg in zip(parallel_dict_index, latest_nodes, trans_cfg["transformers"]):
 
                if "type" in sub_trans_cfg:
                    sub_trans_name = sub_trans_cfg.get("type")
                    ############################### dealling with previous transformer is also parallelTransformer
                    assert sub_trans_name!="parallelTransformer", "parallel is not acceptable in parallelTransfomer"
                    ######################### END

                    #add note pair
                    index = index + 1
                    sub_breakpoint = sub_trans_cfg.pop("breakpoint") if "breakpoint" in sub_trans_cfg else 0
                    description = sub_trans_cfg.pop("description") if "description" in sub_trans_cfg else ""
                    module_type = trans_cfg.pop("module_type") if "module_type" in sub_trans_cfg else ""
                    TRANSFORMERS = get_register(module_type)
                    # print(TRANSFORMERS)
                    transformer = build_transformer_from_cfg(sub_trans_cfg, TRANSFORMERS)
                    transformer.index = index
                    transformer.node_name = f"{str(index)}.{sub_trans_name}"
                    # transformer.__name__
                    transformer.description = description
                    transformer.breakpoint = sub_breakpoint
                    transformer.print_task_result = print_task_result
                    transformers_list.append(transformer)
                    # update nodename2transformer
                    nodename2transformer[transformer.node_name] = transformer
                    # update steps list
                    previous_node, _latest_node = update_node(latest_node, transformer.node_name)
                    parallel_node_pairs[f"process #{str(i)}"]=(previous_node, _latest_node)
                    node_name_set.add(_latest_node)
                    parallel_latest_node.append(_latest_node)


                elif "pipeline" in sub_trans_cfg:
                    transformer, node_name_set, _node_pairs, _latest_node, index, parallel_index, nodename2transformer = recursive_build_pipeline_from_cfg( sub_trans_cfg, 
                                                                                                                                                            node_name_set, 
                                                                                                                                                            [], 
                                                                                                                                                            latest_node, 
                                                                                                                                                            index, 
                                                                                                                                                            parallel_index, 
                                                                                                                                                            nodename2transformer,
                                                                                                                                                            print_task_result)
                    transformers_list.append(transformer)
                    parallel_node_pairs[f"process #{str(i)}"]=_node_pairs
                    parallel_latest_node.append(_latest_node)


            # updaet latest_node                                                
            previous_node, latest_node = update_node(latest_nodes, parallel_latest_node)
            # update node_pairs
            node_pairs.append(parallel_node_pairs)
            # add steps
            # module_type = trans_cfg.pop("module_type") if "module_type" in trans_cfg else ""
            # TRANSFORMERS = get_register(module_type)
            # print(transformers_list)
            parallel_transformer = build_transformer_from_cfg(trans_cfg, 
                                                              FUNCTIONAL_TRANSFORMERS, 
                                                              transformers_list)
            parallel_transformer.breakpoint = breakpoint
            parallel_transformer.print_task_result = print_task_result
            parallel_transformer.node_name = parallel_node_name
            steps.append(parallel_transformer)
            # update nodename2transformer
            nodename2transformer[parallel_node_name] = parallel_transformer


            
        elif trans_name=="pipeline":
            # recursive function help update node_pairs, latest_node and index
            _pipeline, node_name_set, _node_pairs, latest_node, index, parallel_index, nodename2transformer= recursive_build_pipeline_from_cfg( trans_cfg, 
                                                                                                                                                node_name_set, 
                                                                                                                                                [], 
                                                                                                                                                latest_node, 
                                                                                                                                                index, 
                                                                                                                                                parallel_index, 
                                                                                                                                                nodename2transformer,
                                                                                                                                                print_task_result)
            # update steps
            node_pairs.append(_node_pairs)
            steps.append(_pipeline)
        
        else:
            index = index + 1
            node_name = f"{str(index)}.{trans_name}"
            # breakpoint=trans_cfg.pop("breakpoint") if "breakpoint" in trans_cfg else 0
            description = trans_cfg.pop("description") if "description" in trans_cfg else ""
            module_type = trans_cfg.pop("module_type") if "module_type" in trans_cfg else ""
            TRANSFORMERS = get_register(module_type)
            transformer = build_transformer_from_cfg(trans_cfg, TRANSFORMERS)
            transformer.index = index
            transformer.node_name = node_name
            transformer.description = description
            transformer.breakpoint = breakpoint
            transformer.print_task_result = print_task_result
            
            # update steps list
            steps.append(transformer)
            # update nodename2transformer
            nodename2transformer[node_name] = transformer
            # update node_pairs
            if isinstance(latest_node, str):
                node_pairs.append((latest_node, node_name))
            elif isinstance(latest_node, list):
                for item in latest_node:
                    node_pairs.append((item, node_name))
            node_name_set.add(node_name)
            #update latest_node
            latest_node = node_name

    # remove empty node in node pair
    # simple_node_pairs = [item for item in simple_node_pairs if item[0]!=""]
    node_name_set = set([node_name for node_name in node_name_set if node_name!=""])


    return make_pipeline(*steps), node_name_set, node_pairs, latest_node, index, parallel_index, nodename2transformer
            

def create_tranformers_dict(transformers):
    id2transformer = {}
    nodename2transformer = {}
    for transformer in transformers:
        id2transformer[transformer.index] = transformer
        nodename2transformer[transformer.node_name] = transformer

    return id2transformer, nodename2transformer


# when build_pipeline, it will return Pipeline
class Pipeline(sklearn_pipeline):
    def __init__(self, 
                 steps, 
                 *, 
                 memory=None,
                 verbose=False,
                 pipe_start=1,
                 pipe_stop=None):
        
        self.steps = steps
        self.memory = memory
        self.verbose = verbose
        self.pipe_start = pipe_start
        self.pipe_stop = pipe_stop
        
    
    def _islice(self, target_list, start, end):
        total = len(target_list)
        assert end <=total, "end is higher total elements"
        
        selected_elements=None
        
        if start>0 and end ==-1:
            selected_elements = target_list[start-1:]
        elif start<0 and end==-1:
            selected_elements = target_list[start:]
        elif start>0 and end >0:
            selected_elements = target_list[start-1:end]
        elif start<0 and end <0:
            selected_elements = target_list[start:end+1]
        elif start>0 and end <0:
            selected_elements = target_list[start-1:end+1]
            
        return selected_elements

    
    def _iter(self):
        # print("running _iter #########################")
        if self.pipe_stop ==None:
            self.pipe_stop = len(self.steps)
        for idx, (name, trans) in enumerate(self._islice(self.steps, self.pipe_start, self.pipe_stop)):
            yield idx, name, trans
        
        
    def select_transformers(self, pipe_start=None, pipe_stop=None):
        self.pipe_start = pipe_start
        self.pipe_stop = pipe_stop
        new = copy.deepcopy(self)
        
        new.steps = self._islice(self.steps, self.pipe_start, self.pipe_stop)
        
        return new

    def _can_transform(self):
        return self._final_estimator == "passthrough" or hasattr(
            self._final_estimator, "transform"
        )


    def transform(self, X):
        Xt = X
        for _, _, transformer in self._iter():
            Xt = transformer.transform(Xt)
            if hasattr(transformer, "breakpoint"):
                if transformer.breakpoint:
                    # print(transform.node_name, "has breakpoint", transform.breakpoint)
                    datetime_str = get_datetime_str(format="%Y-%m-%d %H:%M:%S")
                    print(f"{datetime_str} stop with breakpoint at {transformer.node_name}") if transformer.print_task_result else ""
                    return Xt
        return Xt
    


class pipeline_settings():
    def __init__(self, cfg):
        self.pipeline_settings = cfg.pop("pipeline_settings") if "pipeline_settings" in cfg else {"print_cfg":0, "print_task_result":0}
    
    def get_pipeline_settings_from_cfg(self):
        
        print_cfg = self.pipeline_settings.pop("print_cfg") if "print_cfg" in self.pipeline_settings else 0
        print_task_result = self.pipeline_settings.pop("print_task_result") if "print_task_result" in self.pipeline_settings else 0

        return print_cfg, print_task_result
        

class DataPipeline:
    """
    usage
    --------------
    pipe = DataPipeline(cfg)
    pipe.transformer()

    UI Editing
    --------------
    self.generate_node_text()
    """
    def __init__(self, cfg):
        #------------------------- pipeline_settings 
        _pipeline_settings = pipeline_settings(cfg)
        self._print_cfg, \
        self._print_task_result = _pipeline_settings.get_pipeline_settings_from_cfg()

        if self._print_cfg:
            print("".join(["-" for i in range(80)]), "pipeline configuration", "".join(["-" for i in range(80)]))
            print(cfg)
            print("".join(["-" for i in range(80)]), "----------------------", "".join(["-" for i in range(80)]))

        self._cfg = cfg
        self._node_name_set=set()
        self._node_pairs = []
        self._latest_node = ""
        self._index = 0
        self._parallel_index=0
        self._dot = None
        self._nodename2transformer={}
        if self._cfg:
            self._pipeline = self.build_pipeline_from_cfg()
            self.steps = self._pipeline.steps
        else:
            self._pipeline=None
            self.steps=[]

        self.limit = 40 #text line limit of characters in node
        # self.nodes_text = init_nodes_text(self._nodename2transformer)

    def __enter__(self, ):
        # tran = paramiko.Transport((self.host, int(self.port)))
        # tran.connect(username=self.username, password=self.password)
        # self.transport = tran

        # try:        
        #     sftp = paramiko.SFTPClient.from_transport(self.transport)
        #     self.sftp = sftp
        #     return self
        # except:
        #     self.transport.close()
            # raise
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        # self.sftp.close()
        # self.transport.close()
        pass

    def build_pipeline_from_cfg(self):

        
        _pipeline, \
        self._node_name_set, \
        self._node_pairs, \
        self._latest_node, \
        index, \
        parallel_index, \
        self._nodename2transformer= recursive_build_pipeline_from_cfg(  self._cfg, 
                                                                        self._node_name_set,
                                                                        self._node_pairs, 
                                                                        self._latest_node, 
                                                                        self._index,
                                                                        self._parallel_index,
                                                                        self._nodename2transformer,
                                                                        self._print_task_result)
        return _pipeline


    def build_digraph(self):
        """
        Tutorial: https://graphviz.readthedocs.io/en/stable/examples.html
        Note: graphviz.Graph have no direction, so no arrowhead

        Variable Dependencies:
            - self._node_pairs
            - self._nodename2transformer
            - self._node_name_set
        """

        # init the digraph object
        self._dot = init_digraph()

        # setting digraph attributes
        self._dot.attr('node', 
                        shape="rect", 
                        style="filled", 
                        fillcolor="skyblue4", 
                        color="white", 
                        fontcolor="white")


        # create node for digraph
        add_dot_node(self, self._nodename2transformer, self._node_name_set)
            
            

        # create edges
        self._pre_dot, \
        self._dot, \
        parallel_index, \
        pipeline_index = recursive_build_digraph(self._dot, self._dot, self._node_pairs, 0, 0)


        return self



    def save_png(self, out_file):
        if out_file:
            ext = "." + out_file.split(".")[-1]
            out_file = out_file.replace(ext, "")
            # print(out_file)
            self._dot.render(out_file).replace("\\", "/")
    

    def output_graph(self):
        with tempfile.NamedTemporaryFile(suffix=".png") as f:
            # print(f.name)
            dirname = os.path.dirname(f.name)
            base_name = os.path.basename(f.name)
            filename = base_name.replace(".png", "")
            self._dot.render(os.path.join(dirname, filename)).replace("\\", "/")
            return f.name
        
        return ""

    def send_api_graph(self):
        with tempfile.NamedTemporaryFile(suffix=".png") as f:
            # print(f.name)
            dirname = os.path.dirname(f.name)
            base_name = os.path.basename(f.name)
            filename = base_name.replace(".png", "")
            self._dot.render(os.path.join(dirname, filename)).replace("\\", "/")
            return send_file(f.name, as_attachment=True, attachment_filename=base_name)

    def transform(self, input):
        return self._pipeline.transform(input)


def recursive_build_node_from_steps(steps, node_name_set, node_pairs, latest_node, index, parallel_index, nodename2transformer):

    for step_tuple in steps:

        trans_name = step_tuple[0]
        transformer = step_tuple[1]


        if isinstance(transformer, parallelTransformer):

            parallel_name = "parallel"
            parallel_index +=1
            parallel_node_name = f"{parallel_name} #{parallel_index}"

            transformers_list = []
            parallel_latest_node = []
            parallel_node_pairs = {}
            latest_nodes = None

            # prepare latest_nodes
            if isinstance(latest_node, str):
                latest_nodes = [latest_node for i in range(len(transformer.transformers))]
            elif isinstance(latest_node, list):
                latest_nodes = latest_node
            else:
                assert False, "latest_node is not str or list"

            parallel_dict_index = [i for i in range(len(latest_nodes))]

            # dealing the transformers inside the parallelTransformer
            for i, latest_node, sub_transformer in zip( parallel_dict_index, 
                                                        latest_nodes, 
                                                        transformer.transformers):

                ###################################### dealing with pipeline
                if isinstance(sub_transformer, DataPipeline):
                    node_name_set, \
                    _node_pairs, \
                    latest_node, \
                    index, \
                    parallel_index, \
                    nodename2transformer = recursive_build_node_from_steps( sub_transformer.steps,
                                                                            node_name_set,
                                                                            node_pairs, 
                                                                            latest_node, 
                                                                            index, 
                                                                            parallel_index, 
                                                                            nodename2transformer)
                    node_pairs.append(_node_pairs)

                else:
                ####################################### dealing with real transformer
                    index = index + 1
                    description = sub_transformer.description if sub_transformer.description else ""
                    sub_transformer.index = index
                    _tuple = _name_estimators([sub_transformer])
                    sub_transformer.node_name = f"{str(index)}.{_tuple[0][0]}"
                    sub_transformer.description = description
                    transformers_list.append(sub_transformer)
                    # update nodename2transformer
                    nodename2transformer[sub_transformer.node_name] = sub_transformer
                    # update steps list
                    previous_node, _latest_node = update_node(latest_node, sub_transformer.node_name)
                    parallel_node_pairs[f"process #{str(i)}"] = (previous_node, _latest_node)
                    node_name_set.add(_latest_node)
                    parallel_latest_node.append(_latest_node)



            ################################################# dealing with the parallelTransformer
            # update latest_node
            previous_node, latest_node = update_node(latest_node, parallel_latest_node)
            # update node_pairs
            node_pairs.append(parallel_node_pairs)
            nodename2transformer[parallel_node_name] = transformer

        elif isinstance(transformer, DataPipeline):
            node_name_set, \
            _node_pairs, \
            latest_node, \
            index, \
            parallel_index, \
            nodename2transformer = recursive_build_node_from_steps( sub_transformer.steps,
                                                                    node_name_set,
                                                                    node_pairs, 
                                                                    latest_node, 
                                                                    index, 
                                                                    parallel_index, 
                                                                    nodename2transformer)
            node_pairs.append(_node_pairs)
        
        else:

            index = index + 1
            description = transformer.description if transformer.description else ""
            transformer.index = index
            transformer.node_name = f"{str(index)}.{trans_name}"
            transformer.description = description
            # update nodename2transformer
            nodename2transformer[transformer.node_name] = transformer
            # update node_pairs
            if isinstance(latest_node, str):
                node_pairs.append((latest_node, transformer.node_name))
            elif isinstance(latest_node, list):
                for item in latest_node:
                    node_pairs.append((item, transformer.node_name))

            # add current node name to the node_name_set
            node_name_set.add(transformer.node_name)

            #update latest_node
            latest_node = transformer.node_name
        
    node_name_set = set([node_name for node_name in node_name_set if node_name!=""])

    return node_name_set, node_pairs, latest_node, index, parallel_index, nodename2transformer

def build_pipeline(*steps):

    
    steps_tuples = _name_estimators(steps)

    self = DataPipeline("")
    # self.limit = 40 #text line limit of characters in node

    self._node_name_set, \
    self._node_pairs, \
    self._latest_node, \
    _, \
    _, \
    self._nodename2transformer = recursive_build_node_from_steps(   steps_tuples,
                                                                    self._node_name_set,
                                                                    self._node_pairs, 
                                                                    self._latest_node, 
                                                                    self._index,
                                                                    self._parallel_index,
                                                                    self._nodename2transformer)

    # self.nodes_text = init_nodes_text(self._nodename2transformer)

    self._pipeline = make_pipeline(*steps)
    
    return self


from .registry import Registry
import pandas as pd
DataFrame = pd.DataFrame
# pd.set_option('display.max_columns', None)
FUNCTIONAL_TRANSFORMERS = Registry("functional_modules")
PANDAS_TRANSFORMERS = Registry("pandas_modules")
JSON_TRANSFORMERS = Registry("json_modules")
SPARK_TRANSFORMERS = Registry("spark_modules")
TEST_TRANSFORMERS = Registry("test_modules")
context_name="task_data"


DB_MAPPERS = Registry("src_db_mappers")
# tenant388_transaction_models =Registry("tenant388_transaction models")
# TB_MODELS = {"tenant388_transaction":tenant388_transaction_models}
DB_MODELS = Registry("db models")




def get_register(module_type):
    # print("module_type:", module_type)
    TRANSFORMERS=None
    # module_type is like "pandas", "json" or "spark"
    if module_type=="functional":
        TRANSFORMERS = FUNCTIONAL_TRANSFORMERS
    elif module_type=="pandas":
        TRANSFORMERS = PANDAS_TRANSFORMERS
    elif module_type=="json":
        TRANSFORMERS = JSON_TRANSFORMERS
    elif module_type=="spark":
        TRANSFORMERS = SPARK_TRANSFORMERS
    elif module_type=="":
        TRANSFORMERS = PANDAS_TRANSFORMERS
        # print(TRANSFORMERS)
    else:
        raise ValueError(f"module_type, {module_type} is not recognized by cetl")

    # print(TRANSFORMERS)

    return TRANSFORMERS



def build_transformer_from_cfg(cfg, registry, parallel_transformers=None):
    args = cfg.copy()
    transformer_type = args.pop("type")
    
    if transformer_type not in registry.module_dict:
        print(registry)
        raise ValueError(f"transformer_type {transformer_type} not exists")

    cls_obj = registry.module_dict[transformer_type]
        
    
    transformer=None
    if transformer_type=="parallelTransformer":
        transformer = cls_obj(parallel_transformers)
        
    else:
        transformer = cls_obj(**args)

    return transformer


from ..utils.base import Base
from ..utils.builder import JSON_TRANSFORMERS
from ..utils.transform_wrapper import transform_wrapper

@JSON_TRANSFORMERS.add()
class dropColumns(Base):
    """
    {"type":"dropColumns", "subset":["EmployeeID"]}
    """
    def __init__(self, subset=None):
        self.subset = subset

    @transform_wrapper
    def transform(self, input_data:dict):
        for record in input_data["json_data"]:
            for field in self.subset:
                if field in record:
                    record.pop(field)
                else:
                    print(f"field {field} not exists")

        return input_data



import sys 
sys.path.append('.')
from cetl import DataPipeline
from cetl.utils.builder import TRANSFORMERS


result=""
cfg = {"pipeline":[{"type":"readCSVAsJson", 
                        "filepath":"/home/clement/data/for_json_pipeline/Orders.csv", 
                        "delimiter":"\t"},
                    {"type":"dropColumns", "subset":["EmployeeID", "ShipperID"]}
        ]}


pipe = DataPipeline(cfg)
print(pipe.steps)
# result=pipe.transform(result)

print(result)

# python3.6 cetl/tests/sample4_test_for_pipeline_v3.py


from setuptools import setup, find_packages
import codecs
import os

readme_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), "cetl")

with codecs.open(os.path.join(readme_dir, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

# Setting up
setup(
    name="cetl",
    version='0.2.0',
    author="Clement",
    author_email="<cheukub@gmail.com>",
    description='A basic data pipeline tools for data engineer to handle the CRM or loyalty data',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=['flask', 'SQLalchemy', 'pyspark', 'pandas'],
    keywords=['python', 'data pipeline', 'pipeline'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: POSIX :: Linux",
    ]
)


from functools import wraps
import time
from .builder import context_name, DataFrame
from datetime import datetime
from .file_mgt import get_datetime_str


def recursive_records_count(count, data):
    if isinstance(data, dict):
        if context_name in data:
            count += len(data[context_name])
        else:
            print("not acceptable context_name")
            # raise ValueError()
            return count

    elif isinstance(data, DataFrame):
        count += data.shape[0]
    elif isinstance(data, str):
        count += 0
    elif isinstance(data, list):
        for item in data:
            count = recursive_records_count(count, item)
    else:
        raise ValueError("currently only accept list, str, json and pandas data type")
    
    return count



def transform_wrapper(func):
    """
    Usage
    ----------------------------
    class Calculator:
        @transform_wrapper
        def calculate_something(self, num):
            total = sum((x for x in range(0, num**2)))
            return total

        def __repr__(self):
            return f'calc_object:{id(self)}'
    Reference
    ---------------------------
    https://dev.to/kcdchennai/python-decorator-to-measure-execution-time-54hk
    """

    @wraps(func)
    def func_wrapper(*args, **kwargs):
        self = args[0]

        # Calculate the execution time
        start_time = time.perf_counter()
        self.start_datetime = get_datetime_str(format="%Y-%m-%d %H:%M:%S")

        # print processing label
        print(f"Processing {self.node_name}, started at {self.start_datetime} ...") if self.print_task_result else ""

        # execute the transform function
        result = func(*args, **kwargs)

        # Calcuate the end time
        end_time = time.perf_counter()
        total_time = end_time - start_time
        self.end_datetime = get_datetime_str(format="%Y-%m-%d %H:%M:%S")
        # print(f'Function {func.__name__}{args} {kwargs} Took {total_time:.4f} seconds')


        # Adding the execution time to the transformer
        self.total_time = total_time
        
        # print(self)
        # parallelTransformer will make it run double times, not solved yet

        #count the number of records in input and output of the func:
        self.total_input = recursive_records_count(0, args[1])
        self.total_output = recursive_records_count(0, result)

        if self.node_name:
            print(f"    - take time {total_time:.4f} seconds, total input records {self.total_input}, total output records {self.total_output}") if self.print_task_result else ""

            # print(f"stop with breakpoint at {transformer.node_name}") if self.print_task_result else ""
        # print(f"{self.node_name}: take time {total_time:.4f} seconds, total input records {self.total_input}, total output records {self.total_output}")

        return result

    return func_wrapper





import sys 
sys.path.append('.')
from cetl import DataPipeline, build_pipeline
from cetl.functional_modules.pass_dataframe import dummyStart
from cetl.functional_modules.parallel_transformer import parallelTransformer
from cetl.pandas_modules.input import generateDataFrame
from cetl.pandas_modules.two2one import unionAll

cfg = {"pipeline":[ {"type":"dummyStart", "module_type":"functional"},
                    {"type":"parallelTransformer", "transformers":[
                        {"type":"generateDataFrame"},
                        {"type":"generateDataFrame"}
                    ]},
                    {"type":"unionAll"}
]}

# pipe = DataPipeline(cfg)
steps = [   dummyStart(), 
            parallelTransformer([   generateDataFrame(), 
                                    generateDataFrame()]), 
            unionAll()]

pipe = build_pipeline(*steps)
df = pipe.transform("")
# print(df)


pipe = pipe.build_digraph()
pipe.save_png("./sample8.png")


# python3.6 cetl/dev_tests/sample8_build_pipeline.py


import re
import inspect
import graphviz

display_parallel = False
display_sub_pipeline = True

def init_digraph():
    return graphviz.Digraph("pipeline graph", format='png')

def update_node(current_node=None, new_node=None):
    return current_node, new_node

def recursive_build_digraph(_predot, _dot, node_pairs, parallel_index, pipeline_index):
    """
    shape reference: https://graphviz.org/doc/info/shapes.html

    format for node_pairs:
    with display_parallel=True
    --------------------
    [   ('', '1.addRowNumber'), 
        ('1.addRowNumber', 'parallel #1', 'parallelTransformer'), 
        {   'process #0': ('parallel #1', '2.format2String'), 
            'process #1': [ ('parallel #1', '3.filterBy'), 
                            ('3.filterBy', 'parallel #2', 'parallelTransformer'), 
                            {   'process #0': ('parallel #2', '4.addRowNumber'), 
                                'process #1': ('parallel #2', '5.passDataFrame')}, 
                            ('4.addRowNumber', '6.toCSV'), 
                            ('5.passDataFrame', '6.toCSV')], 
            'process #2': [('parallel #1', '7.paddingZero')], 
            'process #3': ('parallel #1', '8.dropColumns')}, 
                            ('2.format2String', '9.passDataFrame'), 
                            ('6.toCSV', '9.passDataFrame'), 
                            ('7.paddingZero', '9.passDataFrame'), 
                            ('8.dropColumns', '9.passDataFrame'), 
                            [('9.passDataFrame', '10.passDataFrame'), ('10.passDataFrame', '11.toCSV')]]
    
    with display_parallel=False
    --------------------
    [('', '1.addRowNumber'), 
        {   'process #0': ('1.addRowNumber', '2.format2String'), 
            'process #1': [('1.addRowNumber', '3.filterBy'), {'process #0': ('3.filterBy', '4.addRowNumber'), 
                                                            'process #1': ('3.filterBy', '5.passDataFrame')}, 
                                                            ('4.addRowNumber', '6.toCSV'), ('5.passDataFrame', '6.toCSV')], 
            'process #2': [('1.addRowNumber', '7.paddingZero')], 
            'process #3': ('1.addRowNumber', '8.dropColumns')}, 
        ('2.format2String', '9.passDataFrame'), 
        ('6.toCSV', '9.passDataFrame'), 
        ('7.paddingZero', '9.passDataFrame'), 
        ('8.dropColumns', '9.passDataFrame'), 
        [('9.passDataFrame', '10.passDataFrame'), ('10.passDataFrame', '11.toCSV')]]

    """
    
    for i, sub_node_pairs in enumerate(node_pairs):
        if isinstance(sub_node_pairs, dict):
            num_parallel_pairs = len(sub_node_pairs)
            # make parallelTransformer cluster
            # if display_parallel:
            #     parallel_index = parallel_index + 1 if num_parallel_pairs>1 else parallel_index
            #     _sub_dot = _dot.subgraph(name=f"cluster_{str(parallel_index)}")
            #     _sub_dot.attr(label=f"parallel cluster #{str(parallel_index)}", color="blue")
            # else:
            _sub_dot = _dot
            
            for j, sub_node_pair in enumerate(sub_node_pairs.values()):
                
                if isinstance(sub_node_pair, tuple):
                    
                    _usage_dot_ = _sub_dot  if j!=0 else _dot
                    if sub_node_pair[0]!="":
                        _usage_dot_.edge(sub_node_pair[0], sub_node_pair[1], label="")

                elif isinstance(sub_node_pair, list):
                    
                    # if display_sub_pipeline:
                    #     num_sub_pipeline_pairs = len(sub_node_pair)
                    #     pipeline_index = pipeline_index + 1 if num_sub_pipeline_pairs>1 else pipeline_index
                    #     with _sub_dot.subgraph(name=f"cluster_{str(pipeline_index)}") as _sub_pipeline_dot:
                    #         _sub_pipeline_dot.attr(label=f"pipeline cluster #{str(pipeline_index)}", color="blue")
                    #         _, _, parallel_index, pipeline_index = recursive_build_digraph(_sub_dot, _sub_pipeline_dot, sub_node_pair, parallel_index, pipeline_index)
                    # else:
                    _sub_pipeline_dot = _sub_dot
                    _, _, parallel_index, pipeline_index = recursive_build_digraph(_sub_dot, _sub_pipeline_dot, sub_node_pair, parallel_index, pipeline_index)

        elif isinstance(sub_node_pairs, list):
            num_pipeline_pairs = len(sub_node_pairs)
            # make pipeline cluster
            # if display_sub_pipeline:
            #     pipeline_index = pipeline_index + 1 if num_pipeline_pairs> 1 else pipeline_index
            #     with _dot.subgraph(name=f"cluster_{str(pipeline_index)}") as c:
            #         c.attr(label=f"pipeline cluster #{str(pipeline_index)}", color="blue")
            #         _, _, parallel_index, pipeline_index = recursive_build_digraph(_dot, c, sub_node_pairs, parallel_index, pipeline_index)
            # else:
            # if display_pipeline is False, do not use subgraph
            _, _, parallel_index, pipeline_index = recursive_build_digraph(_dot, _dot, sub_node_pairs, parallel_index, pipeline_index)

        elif isinstance(sub_node_pairs, tuple):
            _usage_dot_ = _dot if i!=0 else _predot
            # make transformer edge
            node_pair = sub_node_pairs
            if node_pair[0]!="":
                _usage_dot_.edge(node_pair[0], node_pair[1], label="") 
                if len(node_pair)>2:
                    if node_pair[2]=="parallelTransformer":
                        _usage_dot_.node(node_pair[1], shape='Msquare')
                else:
                    pass
                    # _dot.node(node_pair[0], shape='rect')
                    # _dot.node(node_pair[1], shape='rect')

    return _predot, _dot, parallel_index, pipeline_index



def get_node_text(nodes_text, node_name):
    return nodes_text[node_name]

def append_node_text(nodes_text, node_name, append_text, limit, apply_limit=False):

    if apply_limit:
        if len(append_text)>limit:
            append_text = append_text[:limit]+"..."
    
    nodes_text[node_name] = nodes_text[node_name]+f"{append_text}\l"

def append_description_node_text(nodes_text, node_name, description, limit):
    """
    \l means new line
    """

    if len(description)>limit:
        description = description[:limit]+"..."
    if description:
        description = f"{description}"

    description = description if description else "no description"
    sep_line = "".join(["-" for i in range(int(len(description)*3/2))])
    # print("##########33", sep_line)

    description_text = f"{description}\l{sep_line}"
    append_node_text(nodes_text, node_name, description_text, limit)



def append_input_node_text(nodes_text, node_name, inputs, limit):
    input_text=""
    for input in inputs:
        # print(input, len(input))
        if len(input)>limit:
            input_text += "        " + input[:limit]+"...\l"
        else:
            input_text += "        " + input+"\l"
    #https://www.digitalocean.com/community/tutorials/python-trim-string-rstrip-lstrip-strip
    input_text = input_text.lstrip()[:-2] if input_text else input_text
    append_node_text(nodes_text, node_name, f"Input: {input_text}", limit)


def append_output_node_text(nodes_text, node_name, output, limit):

    if len(output)>limit:
        output = output[:limit]+"..."
    append_node_text(nodes_text, node_name, f"Output: {output}", limit)


def add_dot_node(DataPipeline, _nodename2transformer, transformers_set):
    self=DataPipeline

    nodes_text={}
    for key in _nodename2transformer.keys():
        nodes_text[key]=""
    
    for nodename in transformers_set:
        if "parallelTransformer" in nodename:
            pass
        elif "parallel" in nodename:
                self._dot.node(nodename)
        else:
            transformer = _nodename2transformer[nodename]
            # get transformer description
            description = transformer.description
            #get the arguments of transformer.transform() as input
            # inputs = [nodename for nodename in inspect.getfullargspec(transformer.transform).args if nodename not in ["self"]]
            # get the return value type
            # https://stackoverflow.com/questions/49560974/inspect-params-and-return-types
            # output = str(inspect.signature(transformer.transform).return_annotation)
            # # "<class 'int'>" extract the 'int'
            # output = re.findall("\<class '+(.*?)\'>",output)
            # output = output[0].split(".")[-1] if output else ""

            ############### description text
            append_description_node_text(nodes_text, nodename, description, self.limit)
            ############### nodename text
            append_node_text(nodes_text, nodename, f"Task ID: {nodename}", self.limit, apply_limit=True)
            ############### Input text
            input = transformer.total_input
            input_textline = f"Input: {input:.0f} records"
            append_node_text(nodes_text, nodename, input_textline, self.limit, apply_limit=True)
            ############### Output text
            output = transformer.total_output
            output_textline = f"Output: {output:.0f} records"
            append_node_text(nodes_text, nodename, output_textline, self.limit, apply_limit=True)
            # get the time of execution
            total_time = transformer.total_time
            total_time_textline = f"Take time: {total_time:.4f} seconds"
            append_node_text(nodes_text, nodename, total_time_textline, self.limit, apply_limit=True)

            # get the start time of execution
            start_datetime = transformer.start_datetime
            start_datetime_textline = f"Start at: {start_datetime}"
            append_node_text(nodes_text, nodename, start_datetime_textline, self.limit, apply_limit=True)

            # get the end time of execution
            end_datetime = transformer.end_datetime
            end_datetime_textline = f"End at: {end_datetime}"
            append_node_text(nodes_text, nodename, end_datetime_textline, self.limit, apply_limit=True)

            self._dot.node(nodename, nodes_text[nodename])


# def create_simple_dot_graph(self):
#     """
#     Tutorial: https://graphviz.readthedocs.io/en/stable/examples.html

#     Variables:
#         - _simple_node_pairs = [("1", "2"), ("2", "3), ("3", )]
#             "1" > "2" > "3"
#         - transformers_set = ["1", "2", "3"]
#     """
#     self._dot = graphviz.Graph("pipeline graph", format='png')
#     # general styel
#     self._dot.attr('node', 
#                     shape="rect", 
#                     style="filled", 
#                     fillcolor="skyblue4", 
#                     color="white", 
#                     fontcolor="white")
#     self._dot.attr('edge',
#                     arrowhead="normal",
#                     arrowsize='1.0')
#     # self._dot.graph_attr.update(directed="true")
#     # self._dot.edge_attr(("directed","true"))
    

#     # add node
#     transformers_list=[]
#     for item in self._simple_node_pairs:
#         transformers_list = transformers_list + list(item)
#     transformers_set = sorted(set(transformers_list))
#     for item in transformers_set:
#         # add node
#         self._dot.node(item, item+"\n")

#     # # add edges
#     for item in self._simple_node_pairs:
#         # add edge
#         self._dot.edge(item[0], item[1], 
#                         label="")
#     return self._dot





import sys 
sys.path.append('.')
from cetl import DataPipeline




from cetl import build_pipeline
from cetl.pandas_modules import generateDataFrame, unionAll
from cetl.functional_modules import dummyStart, parallelTransformer

pipe = build_pipeline(   dummyStart(),
                        parallelTransformer([generateDataFrame(), generateDataFrame()]), 
                        unionAll())
df = pipe.transform("")
print(df)


pipe = pipe.build_digraph()
pipe.save_png("./sample6.png")


# python3.6 cetl/dev_tests/sample6_build_pipeline.py


from datetime import datetime
import fnmatch
import os
import tempfile

def gen_uniq_filename(filename):
        today = datetime.now()
        # conver the datetime to string
        now_datetime_str = datetime.strftime(today, '%Y-%m-%d %H:%M:%S')
        return now_datetime_str

def get_datetime_for_filename(format='%Y%m%d%H%M%S'):
        today = datetime.now()
        # conver the datetime to string
        now_datetime_str = datetime.strftime(today, format)
        return now_datetime_str

def get_temp_name(file_prefix, file_suffix):
        # generate temp outfile
        file_datetime = get_datetime_for_filename()
        temp = tempfile.NamedTemporaryFile(prefix=file_prefix+file_datetime+"_", suffix=file_suffix)
        basename = os.path.basename(temp.name)
        return basename

def get_file_ext(filename):
        split_values = filename.split(".")
        if len(split_values)>1:
                ext = "."+split_values[-1]
        else:
                ext = ""
        return ext

def get_datetime_str(format=None):
    today = datetime.now()
    now_datetime_str = datetime.strftime(today, format)
    return now_datetime_str

